from subprocess import check_output
from bisect import bisect_left
from signal import SIGINT
from re import compile, search
from random import randrange
from os import getenv

from ipykernel.kernelbase import Kernel
from ipykernel.comm import CommManager
from pexpect import replwrap

from .images import extract_image_filenames, display_data_for_image
from .commands import commands


class BiochamKernel(Kernel):
    implementation = 'Biocham'
    implementation_version = '0.1.0'
    language = 'biocham'
    language_info = {
        'name': 'biocham',
        'mimetype': 'text/plain',
        'file_extension': '.bc',
        'pygments_lexer': 'prolog',
        'codemirror_mode': 'biocham',
    }
    banner = check_output(['biocham', '--version']).decode('utf-8')
    language_version = banner.split()[1]
    help_links = [{
        'text': 'Biocham manual',
        'url': 'http://lifeware.inria.fr/biocham4/doc/'
    }]
    shell = None
    command_sep = compile('\.\s+')
    timeout = float(getenv('BIOCHAM_TIMEOUT', '120'))

    def __init__(self, **kwargs):
        Kernel.__init__(self, **kwargs)
        # Comms are Jupyter's communication channels
        # we have to tell our manager that we will handle all three basic
        # types of messages (open, close and real messages)
        self.comm_manager = CommManager(parent=self, kernel=self)
        comm_msg_types = ['comm_open', 'comm_msg', 'comm_close']
        for msg_type in comm_msg_types:
            self.shell_handlers[msg_type] = getattr(self.comm_manager,
                                                    msg_type)
        self._start_biocham()

    def _start_biocham(self):
        self.biocham = replwrap.REPLWrapper(
            'biocham --jupyter',
            'biocham: ',
            None,
            continuation_prompt='|:'
        )
        # log = open('foo.log', 'w')
        # self.biocham.child.logfile = log

    def do_execute(self, code, silent, store_history=True,
                   user_expressions=None, allow_stdin=False):
        payload = []

        # inform IPython users that biocham does not support shell magics
        if code.startswith('!'):
            output = 'No system magics available.'
            image_filenames = []
        # handle biocham magic
        elif code.startswith('%') and not code.startswith('% ') and \
                '\n' not in code:
            output, payload = self._do_magic(code[1:])
            image_filenames = []
        # else, run code
        else:
            code = '\n'.join([line.strip() for line in code.splitlines() if
                              line.strip()])
            try:
                output = self.biocham.run_command(code,
                                                  timeout=self.timeout)
            except:
                self.biocham.child.kill(SIGINT)
                self.biocham.run_command('a')
                output = 'non-terminated command'
            image_filenames, output = extract_image_filenames(output)

        if not silent:
            stream_content = {'name': 'stdout', 'text': output}
            self.send_response(self.iopub_socket, 'stream', stream_content)

        for filename in image_filenames:
            try:
                data = display_data_for_image(filename, self.biocham.child.pid)
            except ValueError as e:
                message = {'name': 'stdout', 'text': str(e)}
                self.send_response(self.iopub_socket, 'stream', message)
            else:
                self.send_response(self.iopub_socket, 'display_data', data)

        return {
            'status': 'ok',
            # The base class increments the execution count
            'execution_count': self.execution_count,
            'payload': payload,
            'user_expressions': {},
        }

    def do_complete(self, code, cursor_pos):
        current = code[:cursor_pos]
        prefix = self.command_sep.split(current)[-1]
        return {
            'status': 'ok',
            'cursor_start': cursor_pos - len(prefix),
            'cursor_end': cursor_pos,
            'matches': self._find(prefix),
        }

    def _find(self, prefix):
        matches = []
        index = bisect_left(commands, prefix)
        while commands[index].startswith(prefix):
            matches.append(commands[index])
            index += 1
        return matches

    def _do_magic(self, magic):
        if magic.startswith('load '):
            bcfile = magic[5:].split()[0]
            if not bcfile.endswith('.bc') or bcfile.startswith('.'):
                return 'Cannot read {}'.format(bcfile), []
            output = 'loading {}'.format(bcfile)
            with open(bcfile, 'r') as bcinput:
                lines = bcinput.read().split('.\n')
            # create a cell after current cell for each dotted command
            payload = [{
                'source': 'set_next_input',
                'text': line.strip() + '.',
                'replace': False,
            } for line in reversed(lines) if line.strip()]
        elif magic.startswith('slider '):
            output = ''
            payload = []
            parameters = magic[7:].split()
            comm_id = 'comm_' + str(randrange(1024))

            list_parameters = self.biocham.run_command('list_parameters.')
            current_val = []
            max = []
            for p in parameters:
                cur = search('] parameter\({}=([^)]*)\)'.format(p),
                             list_parameters)
                if cur is None or float(cur.group(1))==0 :
                    current_val.append(0)
                    max.append(10)
                else:
                    current_val.append(cur.group(1))
                    max.append(float(cur.group(1)) * 10)

            # register this slider on Kernel side, with callback handle_comm
            self.comm_manager.register_target(comm_id, handle_comm)
            self.send_response(
                self.iopub_socket,
                'display_data',
                {
                    # the data contains HTML+JS to create the slider
                    # and send a message on value change
                    'data': {
                        'text/html': ''.join([
                            '''
            <label style="float: left; padding-left: 10px; padding-right: 0px; margin-right: 0px;">{param}</label>
            <label id="{param}_val_{id}"
                style="float: left; margin-left: 0px; padding-left: 0px;">
                ={current}</label>
            <div class="slider" id="slider_{param}_{id}"
                style="width: 200px; float: left; margin-top: 4px;"></div>
                            '''.format(param=param, current=current,
                                       id=comm_id) for
                            param, current in zip(parameters, current_val)]) +
                        '''
            <div id="{comm}_out" style="clear: both;"></div>
            <script>
                function add_{comm}() {{
                    {comm} = Jupyter.notebook.kernel.comm_manager.new_comm(
                        '{comm}',
                        {{}}
                    );
                    {comm}.on_msg(function(msg) {{
                        console.log(msg.content.data);
                        $('#{comm}_out').html(msg.content.data);
                    }});'''.format(comm=comm_id) + ''.join(['''
                    $('#slider_{param}_{id}').slider({{
                        value: {current},
                        max: {max},
                        step: {step},
                        change: function(event, ui) {{
                            {id}.send({{name: "{param}", value: ui.value}});
                        }},
                        slide: function(event, ui) {{
                            $('#{param}_val_{id}').html('='+ui.value);
                        }}
                    }});
                '''.format(param=param, id=comm_id, current=current,
                           max=max_val, step=max_val/100) for param, current,
                           max_val in
                           zip(parameters, current_val, max)]) + '''
                }};
                if (typeof Jupyter.notebook.kernel !== "undefined" &&
                    Jupyter.notebook.kernel !== null) {{
                    add_{id}();
                }} else {{
                    require('base/js/events').on('kernel_ready.Kernel',
                                                function(event, info) {{
                        add_{id}();
                    }});
                }}
            </script>'''.format(id=comm_id)
                    },
                    'metadata': {}
                }
            )
        else:
            output = '''The available magics are:
    %lsmagic\tLists available magics
    %load [file.bc]\tImport a biocham file as a notebook cell
    %slider <parameter> [parameter parameter]\tCreates a slider to change \
given parameter(s) in default simulation
'''
            payload = []
        return output, payload


def handle_comm(comm, msg):
    @comm.on_msg
    def _recv(msg):
        # set parameter value to what we receive
        code = 'parameter({param}={value}).'.format(
            param=msg['content']['data']['name'],
            value=msg['content']['data']['value']
        )
        comm.kernel.biocham.run_command(code)
        # run a simulation with default options
        comm.kernel.biocham.run_command('numerical_simulation.',
                                        timeout=comm.kernel.timeout)
        # plot and get corresponding HTML output
        image_filenames, output = extract_image_filenames(
            comm.kernel.biocham.run_command('plot.'))
        data = display_data_for_image(image_filenames[0],
                                      comm.kernel.biocham.child.pid)
        # send that output to be put in the document
        comm.send(data['data']['text/html'])
