:- module(
  markov,
  [
    % API
    stochastic_simulation/4
  ]
).


:- dynamic(is_boolean/0).
:- dynamic(last_time/1).  % last written state's time

:- devcom('
stochastic_simulation(Until, Propensities, Time, Boolean).').

stochastic_simulation(Until, Propensities, Time, Boolean) :-
  (
    Boolean == true
  ->
    assertz(is_boolean)
  ;
    retractall(is_boolean)
  ),
  get_initial_numbers(InitialNumbers),
  get_conditions_and_actions(Conditions, Actions),
  list_to_assoc(['Time'-0 | InitialNumbers], InitState),
  init_state(InitState),
  retractall(last_time(_)),
  assertz(last_time(0.0)),
  markov_run(InitState, Conditions, Actions, Until, Propensities, Time),
  make_state_table.


% hack to force initial state from outside without items
:- dynamic(initial_state/1).


get_initial_numbers(InitialNumbers) :-
  initial_state(InitialNumbers),
  !.

get_initial_numbers(InitialNumbers) :-
  get_option(stochastic_conversion, Rate),
  enumerate_molecules(Molecules),
  findall(
    Molecule-Initial,
    (
      member(Molecule, Molecules),
      get_initial_concentration(Molecule, InitialC),
      (
        is_boolean
      ->
        Initial is round(sign(InitialC))
      ;
        Initial is round(InitialC*Rate)
      )
    ),
    InitialNumbers
  ).


:- dynamic(cache_conditions/2).


get_conditions_and_actions(Conditions, Actions) :-
  nomake_state_table,
  cache_conditions(Conditions, Actions),
  !.


get_conditions_and_actions(Conditions, Actions) :-
  findall(
    (Expr, Reactants, Products),
    (
        item([kind: reaction, item: Item]),
        reaction(
          Item,
          [
            kinetics: Kinetics,
            reactants: Reactants,
            inhibitors: Inhibitors,
            products: Products
          ]
        ),
        kinetics(Reactants, Inhibitors, Kinetics, Expr)
%        kinetics(Reactants, Inhibitors, Kinetics, KExpr),
%        inhibitors_to_kinetics(Inhibitors, Test),
%        Expr = Test * KExpr
      ;
        item([kind: influence, item: Item]),
        influence(
          Item,
          [
            force: Force,
            positive_inputs: Inputs,
            negative_inputs: Inhibitors,
            sign: Sign,
            output: Output
          ]
        ),
        (
          Sign = '+'
        ->
          Products = [1*Output],
          Reactants = [],
          (
            is_boolean
          ->
            AllInputs = [1 - Output | Inputs]
          ;
            AllInputs = Inputs
          )
        ;
          Products = [],
          Reactants = [1*Output],
          % for inhibitions add output to inputs for default force MA(1)
          AllInputs = [Output | Inputs]
        ),
        maplist(influence_editor:add_coefficient, AllInputs, InputsWithCoeff),
        kinetics(InputsWithCoeff, Inhibitors, Force, Expr)
        %inhibitors_to_kinetics(Inhibitors, Test),
        %Expr = Test * KExpr
    ),
    AllStuff
  ),
  enumerate_molecules(Molecules),
  triplets_to_conditions_and_actions(AllStuff, Conditions, Actions, Molecules),
  (
    nomake_state_table
  ->
    assertz(cache_conditions(Conditions, Actions))
  ;
    true
  ).


inhibitors_to_kinetics([], 1) :-
  !.

inhibitors_to_kinetics(Inhibitors, Expr) :-
  foldl(is_plus, Inhibitors, 0, Test),
  % FIXME can be wrong if some number of molecules is negative
  Expr = 1 - sign(Test).


is_plus(X, Y, X + Y).


triplets_to_conditions_and_actions([], [], [], _).

triplets_to_conditions_and_actions(
  [(K, Reactants, Products) | Stuff],
  [KK | Conditions],
  [Action | Actions],
  Molecules
) :-
  normalize_condition(K, KK, Molecules),
  reaction_editor:simplify_catalyst(
    Reactants, Products,
    SReactants, _, SProducts
  ),
  stoichiometry_to_action(SReactants, '-', [], A1),
  stoichiometry_to_action(SProducts, '+', A1, Action),
  triplets_to_conditions_and_actions(Stuff, Conditions, Actions, Molecules).


normalize_condition(Cond, SCond, _Molecules) :-
  is_boolean,
  !,
  if_molecule_divide(1, [], Cond, NCond),
  simplify(NCond, SCond).

normalize_condition(Cond, SCond, Molecules) :-
  get_option(stochastic_conversion, Rate),
  if_molecule_divide(Rate, Molecules, Cond, NCond),
  simplify(Rate*NCond, SCond).


if_molecule_divide(Rate, _Molecules, [M], [M]/Rate) :-
  !.

if_molecule_divide(Rate, Molecules, M, MM) :-
  atom(M),
  !,
  (
    member(M, Molecules)
  ->
    MM = M/Rate
  ;
    parameter_value(M, MM)
  ->
    true
  ;
    MM = M
  ).

if_molecule_divide(Rate, Molecules, M, MM) :-
  term_morphism(markov:if_molecule_divide(Rate, Molecules), M, MM).


stoichiometry_to_action([], _, A, A).

stoichiometry_to_action([_*M | L], '+', A, B) :-
  is_boolean,
  !,
  stoichiometry_to_action(L, '-', [M-1 | A], B).

stoichiometry_to_action([_*M | L], '-', A, B) :-
  is_boolean,
  !,
  stoichiometry_to_action(L, '-', [M-0 | A], B).

stoichiometry_to_action([N*M | L], Op, A, B) :-
  Next =.. [Op, M, N],
  stoichiometry_to_action(L, Op, [M-Next | A], B).


markov_run(State, Conditions, Actions, Until, PropensitiesType, TimeType) :-
  maplist(eval_in_state(State), Conditions, Propensities),
  (
    PropensitiesType = 'boolean'
  ->
    maplist(sign, Propensities, TypedPropensities)
  ;
    TypedPropensities = Propensities
  ),
  sum_list(TypedPropensities, Alpha),
  (
    Alpha =:= 0
  ->
    NewTime = Until,
    NewState = State
  ;
    (
      TimeType = 'discrete'
    ->
      DeltaT = 1
    ;
      DeltaT is log(1/random_float)/Alpha
    ),
    eval_in_state(State, 'Time', Time),
    NewTime is Time + DeltaT,
    Bound is Alpha*random_float,
    % write(select_action(Bound, Actions, Propensities)),nl,
    select_action(Bound, Actions, TypedPropensities, Action),
    set_state(Action, State, NewState)
  ),
  % FIXME handle events -> care with parameters put directly in Conds
  set_state(['Time'-NewTime], NewState, NextState),
  retract(last_time(LastTime)),
  (
    % FIXME this is a very rough sampling
    ((NewTime >= Until) ; (NewTime - LastTime >= min(1, Until/1000.0)))
  ->
    write_state(NextState),
    assertz(last_time(NewTime))
  ;
    assertz(last_time(LastTime))
  ),
  (
    NewTime >= Until
  ->
    true
  ;
    markov_run(NextState, Conditions, Actions, Until, PropensitiesType,
      TimeType)
  ).


eval_in_state(State, [Object], Value) :-
  !,
  get_assoc(Object, State, Value).

eval_in_state(State, Object, Value) :-
  atom(Object),
  !,
  get_assoc(Object, State, Value).

eval_in_state(State, Expr, Value) :-
  term_morphism(markov:eval_in_state(State), Expr, Term),
  % FIXME should hande conditional expressions and functions
  Value is Term.


select_action(Bound, [Action | _], [Propensity | _], Action) :-
  Bound < Propensity,
  !.

select_action(Bound, [_ | Actions], [P | Propensities], Action) :-
  NextBound is Bound - P,
  select_action(NextBound, Actions, Propensities, Action).


set_state([], State, State).

set_state([Key-Value | Actions], State, NewState) :-
  eval_in_state(State, Value, Eval),
  put_assoc(Key, State, Eval, NextState),
  set_state(Actions, NextState, NewState).


:- dynamic(state/1).


init_state(State) :-
  retractall(state(_)),
  del_assoc('Time', State, Time, Rest),
  assoc_to_keys(Rest, Keys),
  assertz(state(['#Time' | Keys])),
  assoc_to_values(Rest, Values),
  assertz(state([Time | Values])).


write_state(State) :-
  del_assoc('Time', State, Time, Rest),
  assoc_to_values(Rest, Values),
  assertz(state([Time | Values])).


% allow avoiding tables alltogether
:- dynamic(nomake_state_table/0).


make_state_table :-
  nomake_state_table,
  !.

make_state_table :-
  findall(
    Row,
    (
      state(State),
      Row =.. [row | State]
    ),
    Table
  ),
  add_table(stochastic_simulation, Table).


export_state(Filename) :-
  automatic_suffix(Filename, '.csv', write, FilenameCsv),
  working_directory(Dir, Dir),
  absolute_file_name(FilenameCsv, AbsoluteFilenameCsv, [relative_to(Dir)]),
  retract(state(Header)),
  !,
  HeaderRow =.. [row | Header],
  setup_call_cleanup(
    open(AbsoluteFilenameCsv, write, Out),
    (
      % properly quote header line
      csv_write_stream(Out, [HeaderRow], []),
      forall(
        state(State),
        (
          atomic_list_concat(State, ', ', Atom),
          write(Out, Atom),
          nl(Out)
        )
      )
    ),
    (
      asserta(state(Header)),
      close(Out)
    )
  ).
