:- module(
  search,
  [
    search_parameters/3
  ]
).


% for linting purposes
:- use_module(objects).
:- use_module(biocham).


:- initial('option(cmaes_init_center: no)').


search_parameters(Formula, Parameters, Objective) :-
  biocham_command,
  type(Formula, foltl),
  type(Parameters, [parameter_between_interval]),
  type(Objective, [variable_objective]),
  option(time, time, _Time, 'time horizon of the numerical integration'),
  option(method, method, _Method, 'method for the numerical solver'),
  option(
    error_epsilon_absolute, number, _ErrorEpsilonAbsolute,
    'absolute error for the numerical solver'
  ),
  option(
    error_epsilon_relative, number, _ErrorEpsilonRelative,
    'relative error for the numerical solver'
  ),
  option(
    initial_step_size, number, _InitialStepSize,
    'initial step size for the numerical solver'
  ),
  option(
    maximum_step_size, number, _MaximumStepSize,
    'maximum step size for the numerical solver, as a fraction of time'
  ),
  option(
    precision, number, _Precision,
    'precision for the numerical solver'
  ),
  option(
    cmaes_init_center, yesno, _CmaesInit,
    'initialize parameter values at the center of their domain instead of
    their current value'
  ),
  doc('
    tries to satisfy a FOLTL(Rlin) constraint by varying the parameters listed
    in \\argument{Parameters}.
    This search command uses the stochastic optimization procedure CMA-ES
    (covariance matrix adaptation evolution strategy)
    with the continuous satisfaction degree of a FO-LTL(Rlin) property as
    fitness function.
    This satisfaction degree is defined by the distance between the validity
    domain of a FOLTL(Rlin) formula with free variables (as given by the command
    domains) and the objective values given in \\command{Objective}.
\\clearmodel
\\begin{example}
\\trace{
biocham: k for a => b. present(a,10).
biocham: search_parameters(G(Time < T => a > b), [0 <= k <= 2], [T -> 5]).
biocham: search_parameters(G(Time < T => a > b), [0 <= k <= 2], [T -> 5]).
biocham: search_parameters(G(Time < T => a > b), [0 <= k <= 2], [T -> 5]).
}
\\end{example}
  '),
  with_current_ode_system(
    with_clean(
      [
        numerical_simulation:variable/2,
        numerical_simulation:equation/2,
        numerical_simulation:parameter_index/2,
        numerical_simulation:conditional_event/2
      ],
      search_parameters(Formula, Parameters, Objective, Result)
    )
  ),
  set_parameters(Result),
  find_parameter_ids(Parameters, ParameterIds),
  list_ids(ParameterIds).



search_parameters(Formula, Parameters, Objective, Result) :-
  GslCFilename = 'ode.inc',
  DomainCFilename = 'check.inc',
  SearchParametersCFilename = 'search_parameters.inc',
  ExecutableFilename = 'search',
  ensure_parameters(Parameters),
  expand_formula(Formula, ExpandedFormula),
  prepare_numerical_simulation_options(Options),
  with_output_to_file(GslCFilename, write_gsl(Options)),
  with_clean(
    [foltl:column/2],
    (
      populate_field_columns(Options),
      with_output_to_file(DomainCFilename, generate_domain_cpp(ExpandedFormula))
    )
  ),
  with_output_to_file(
    SearchParametersCFilename,
    (
      generate_random_seed,
      generate_search_parameters(Parameters),
      generate_objective(Objective),
      generate_add_row(Options)
    )
  ),
  compile_search_cpp_program(ExecutableFilename),
  call_subprocess(
    ExecutableFilename, [], [stdout(pipe(ResultStream))]
  ),
  read_parameter_values(Parameters, ResultStream, Result),
  read_satisfaction_degree(ResultStream, Degree),
  format('Best satisfaction degree: ~p~n', [Degree]),
  (
    have_to_delete_temporary_files
  ->
    delete_file(GslCFilename),
    delete_file(DomainCFilename),
    delete_file(SearchParametersCFilename),
    delete_file(ExecutableFilename)
  ;
    true
  ).


generate_random_seed :-
  (
    toplevel:random_seed(Seed)
  ->
    NewSeed is Seed + 1
  ;
    NewSeed = 0
  ),
  format('#define SEED ~d~n', [NewSeed]).


generate_search_parameters(Parameters) :-
  get_option(cmaes_init_center, Center),
  length(Parameters, ParameterCount),
  format('#define DIMENSION ~d\n', [ParameterCount]),
  write('\c
static const struct search_parameter search_parameters[DIMENSION] = {\n'),
  \+ (
    member(Min <= Parameter <= Max, Parameters),
    \+ (
      numerical_simulation:parameter_index(Parameter, ParameterIndex),
      (
        Center == yes
      ->
        Value is (Min + Max)/2
      ;
        parameter_value(Parameter, Value)
      ),
      format('    {~f, ~d, ~f, ~f},\n', [Min, ParameterIndex, Max, Value])
    )
  ),
  write('};\n').


generate_add_row(Options) :-
  memberchk(fields: Fields, Options),
  write('\c
static void
add_row(Row &row, gsl_solver_state &state)
{
    double t = state.t;
    double *p = state.solver->p;
    double *x = state.x;
    bool *e = state.e;
'),
  \+ (
    member(_: Field, Fields),
    \+ (
      write('    row.push_back('),
      write_field(Field),
      write(');\n')
    )
  ),
  write('};\n').


compile_search_cpp_program(ExecutableFilename) :-
  gsl_compile_options(GslCompileOptions),
  ppl_compile_options(PplCompileOptions),
  append(GslCompileOptions, PplCompileOptions, AllCompileOptions),
  compile_cpp_program(
    ['search_parameters.cc', 'gsl_solver.o', 'cmaes.o'],
    AllCompileOptions,
    ExecutableFilename
  ).


ensure_parameters(Parameters) :-
  member(Min <= Parameter <= _Max, Parameters),
  (
    parameter_value(Parameter, _)
  ->
    true
  ;
    set_parameter(Parameter, Min)
  ).


find_parameter_ids([], []).

find_parameter_ids([_Min <= Parameter <= _Max | Tail], [Id | IdTail]) :-
  find_item([key: Parameter, id: Id]),
  find_parameter_ids(Tail, IdTail).


populate_field_columns(Options) :-
  memberchk(fields: Fields, Options),
  \+ (
    nth0(Index, Fields, Field: _),
    \+ (
      assertz(foltl:column(Field, Index))
    )
  ).



read_parameter_values([], _ResultStream, []).

read_parameter_values(
  [_Min <= Parameter <= _Max | ParameterTail], ResultStream,
  [Parameter = Value | ResultTail]
) :-
  read_line_to_codes(ResultStream, ValueCode),
  number_codes(Value, ValueCode),
  read_parameter_values(ParameterTail, ResultStream, ResultTail).


read_satisfaction_degree(Stream, Degree) :-
  read_line_to_codes(Stream, Codes),
  number_codes(Degree, Codes).
